#include <ctime>
#include <cmath>
#include <algorithm>
#include <fstream>
#include <iomanip>

#include <BWTA.h>
#include <BWAPI.h>

#include <RL/Matrix.h>
#include <RL/NeuronNetwork.h>
#include <RL/RLAgent.h>

#include "MacroAIModule.h"

#define sig(x) (1 / (1 + exp(-(x))))

namespace AgentSettings {
  const int HiddenNeuronCount = 15;
  const double ExplorationRate = 0.0375;
  const double DiscountFactor = 0.9999;
  const double TimeCoeficient = 0.35;
  const double LearningRate = 0.05;
}

namespace StateFields {
  const int UnusedSupply = 0;
  const int ProbeCount = 1;

  const int NexusCount = 2;
  const int ForgeCount = 3;
  const int CannonPylonRatio = 4;
  
  const int EnemyWorkerCount = 5;
  const int EnemyLowLevelUnitCount = 6;
  const int EnemyHightLevelUnitCount = 7;

  const int EnemyTurretCount = 8;
  const int EnemyLowLevelProducerCount = 9;
  const int EnemyHighLevelProducerCount = 10;
  const int EnemyTechBuildingCount = 11;

  const int count = 12;
};

string StateFieldNames[] = {
  "UnusedSupply",
  "ProbeCount",

  "NexusCount",
  "ForgeCount",
  "CannonPylonRatio",
  
  "EnemyWorkerCount",
  "EnemyLowLevelUnitCount",
  "EnemyHightLevelUnitCount",

  "EnemyTurretCount",
  "EnemyLowLevelProducerCount",
  "EnemyHighLevelProducerCount",
  "EnemyTechBuildingCount"
};

namespace Actions {
  const int TrainProbe = 0;
  
  const int BuildBasePylon = 1;
  const int BuildForge = 2;

  const int BuildAttackPylon = 3;
  const int BuildAttackCannon = 4;

  const int Wait = 5;

  const int none = -1;
  const int count = 6;
};

string ActionNames[] = {
  "TrainProbe",
  
  "BuildBasePylon",
  "BuildForge",
  
  "BuildAttackPylon",
  "BuildAttackCannon",

  "Wait"
};

namespace UnitTypeClasses {
  set<int> lowLevelUnitIDs;
  set<int> hightLevelUnitIDs;
  set<int> turretIDs;
  set<int> lowLevelProducerIDs;
  set<int> highLevelProducerIDs;
  set<int> techBuildingIDs;
  set<int> vespeneGeyserIDs;
};

namespace UnitTypeIDs {
  const int ProtossProbe = UnitTypes::Protoss_Probe.getID();
  const int ProtossPylon = UnitTypes::Protoss_Pylon.getID();
  const int ProtossNexus = UnitTypes::Protoss_Nexus.getID();
  const int ProtossForge = UnitTypes::Protoss_Forge.getID();
  const int ProtossPhotonCannon = UnitTypes::Protoss_Photon_Cannon.getID();
}

void initUnitTypeClasses() {
  // initialize LOW LEVEL UNIT type class
  UnitTypeClasses::lowLevelUnitIDs.insert(UnitTypes::Protoss_Zealot.getID());
  UnitTypeClasses::lowLevelUnitIDs.insert(UnitTypes::Protoss_Dragoon.getID());
  UnitTypeClasses::lowLevelUnitIDs.insert(UnitTypes::Terran_Marine.getID());
  UnitTypeClasses::lowLevelUnitIDs.insert(UnitTypes::Terran_Firebat.getID());
  UnitTypeClasses::lowLevelUnitIDs.insert(UnitTypes::Terran_Medic.getID());

  // initialize HIGH LEVEL UNIT type class
  UnitTypeClasses::hightLevelUnitIDs.insert(UnitTypes::Protoss_High_Templar.getID());
  UnitTypeClasses::hightLevelUnitIDs.insert(UnitTypes::Protoss_Dark_Templar.getID());
  UnitTypeClasses::hightLevelUnitIDs.insert(UnitTypes::Protoss_Reaver.getID());
  UnitTypeClasses::hightLevelUnitIDs.insert(UnitTypes::Protoss_Scout.getID());
  UnitTypeClasses::hightLevelUnitIDs.insert(UnitTypes::Protoss_Carrier.getID());
  UnitTypeClasses::hightLevelUnitIDs.insert(UnitTypes::Protoss_Arbiter.getID());
  UnitTypeClasses::hightLevelUnitIDs.insert(UnitTypes::Protoss_Archon.getID());
  UnitTypeClasses::hightLevelUnitIDs.insert(UnitTypes::Protoss_Dark_Archon.getID());
  UnitTypeClasses::hightLevelUnitIDs.insert(UnitTypes::Terran_Ghost.getID());
  UnitTypeClasses::hightLevelUnitIDs.insert(UnitTypes::Terran_Vulture.getID());
  UnitTypeClasses::hightLevelUnitIDs.insert(UnitTypes::Terran_Siege_Tank_Siege_Mode.getID());
  UnitTypeClasses::hightLevelUnitIDs.insert(UnitTypes::Terran_Siege_Tank_Tank_Mode.getID());
  UnitTypeClasses::hightLevelUnitIDs.insert(UnitTypes::Terran_Goliath.getID());
  UnitTypeClasses::hightLevelUnitIDs.insert(UnitTypes::Terran_Wraith.getID());
  UnitTypeClasses::hightLevelUnitIDs.insert(UnitTypes::Terran_Battlecruiser.getID());
  UnitTypeClasses::hightLevelUnitIDs.insert(UnitTypes::Terran_Science_Vessel.getID());
  
  // initialize TURRRET type class
  UnitTypeClasses::turretIDs.insert(UnitTypes::Protoss_Photon_Cannon.getID());
  UnitTypeClasses::turretIDs.insert(UnitTypes::Terran_Bunker.getID());

  // initialize LOW LEVEL PRODUCER type class
  UnitTypeClasses::lowLevelProducerIDs.insert(UnitTypes::Protoss_Gateway.getID());
  UnitTypeClasses::lowLevelProducerIDs.insert(UnitTypes::Terran_Barracks.getID());

  // initialize HIGH LEVEL PRODUCER type class
  UnitTypeClasses::highLevelProducerIDs.insert(UnitTypes::Protoss_Robotics_Facility.getID());
  UnitTypeClasses::highLevelProducerIDs.insert(UnitTypes::Protoss_Stargate.getID());
  UnitTypeClasses::highLevelProducerIDs.insert(UnitTypes::Terran_Factory.getID());
  UnitTypeClasses::highLevelProducerIDs.insert(UnitTypes::Terran_Starport.getID());

  // initialize TECH BUILDING type class
  UnitTypeClasses::techBuildingIDs.insert(UnitTypes::Protoss_Forge.getID());
  UnitTypeClasses::techBuildingIDs.insert(UnitTypes::Protoss_Cybernetics_Core.getID());
  UnitTypeClasses::techBuildingIDs.insert(UnitTypes::Protoss_Citadel_of_Adun.getID());
  UnitTypeClasses::techBuildingIDs.insert(UnitTypes::Protoss_Templar_Archives.getID());
  UnitTypeClasses::techBuildingIDs.insert(UnitTypes::Protoss_Fleet_Beacon.getID());
  UnitTypeClasses::techBuildingIDs.insert(UnitTypes::Protoss_Arbiter_Tribunal.getID());
  UnitTypeClasses::techBuildingIDs.insert(UnitTypes::Protoss_Robotics_Support_Bay.getID());
  UnitTypeClasses::techBuildingIDs.insert(UnitTypes::Protoss_Observatory.getID());
  UnitTypeClasses::techBuildingIDs.insert(UnitTypes::Terran_Comsat_Station.getID());
  UnitTypeClasses::techBuildingIDs.insert(UnitTypes::Terran_Nuclear_Silo.getID());
  UnitTypeClasses::techBuildingIDs.insert(UnitTypes::Terran_Engineering_Bay.getID());
  UnitTypeClasses::techBuildingIDs.insert(UnitTypes::Terran_Academy.getID());
  UnitTypeClasses::techBuildingIDs.insert(UnitTypes::Terran_Machine_Shop.getID());
  UnitTypeClasses::techBuildingIDs.insert(UnitTypes::Terran_Armory.getID());
  UnitTypeClasses::techBuildingIDs.insert(UnitTypes::Terran_Control_Tower.getID());
  UnitTypeClasses::techBuildingIDs.insert(UnitTypes::Terran_Science_Facility.getID());
  UnitTypeClasses::techBuildingIDs.insert(UnitTypes::Terran_Physics_Lab.getID());
  UnitTypeClasses::techBuildingIDs.insert(UnitTypes::Terran_Covert_Ops.getID());

  // initialize VESPENE GEYSER type class
  UnitTypeClasses::vespeneGeyserIDs.insert(UnitTypes::Resource_Vespene_Geyser.getID());
  UnitTypeClasses::vespeneGeyserIDs.insert(UnitTypes::Protoss_Assimilator.getID());
  UnitTypeClasses::vespeneGeyserIDs.insert(UnitTypes::Terran_Refinery.getID());
}

void MacroAIModule::onStart() {
  srand((unsigned)time(NULL));

  initUnitTypeClasses();

  Broodwar->sendText("black sheep wall");

  BWTA::readMap();
  BWTA::analyze();

  Broodwar->enableFlag(Flag::UserInput);
  Broodwar->setLocalSpeed(this->defaultSpeed);
  Broodwar->setGUI(false);
  
  this->player = Broodwar->self();
  this->enemyPlayer = Broodwar->enemy();

  this->incompleteUnits.clear();

  this->probes.clear();
  this->baseBuilderProbe = NULL;
  this->attackBuilderProbe = NULL;

  this->nexus = NULL;
  this->basePylon = NULL;
  this->forge = NULL;

  this->attackPylons.clear();
  this->attackCannons.clear();

  for (set<Unit*>::const_iterator itUnit = this->player->getUnits().begin(); itUnit != this->player->getUnits().end(); ++itUnit) {
    if ((*itUnit)->getType().getID() == UnitTypeIDs::ProtossProbe) {
      this->probes.insert(pair<int, Unit*>((*itUnit)->getID(), *itUnit));
    } else if ((*itUnit)->getType().getID() == UnitTypeIDs::ProtossNexus) {
      this->nexus = *itUnit;
    }
  }

  this->baseBuildInProgress = false;
  this->buildableBaseTiles.clear();

  this->attackBuildInProgress = false;
  this->buildableHostileTiles.clear();

  this->stronglyDefendedTiles.clear();
  this->weaklyDefendedTiles.clear();

  this->baseAreaTiles.clear();
  this->baseAreaSize = this->restrictedBaseBuildArea + this->areaExpandStep;
  this->canBuildBasePylon = true;
  this->canBuildForge = false;

  this->hostileAreaTiles.clear();
  this->hostileAreaSize = this->restrictedFirstPylonArea + this->areaExpandStep;
  this->canBuildAttackPylon = false;
  this->canBuildAttackCannon = false;

  this->startLocationTilePosition = this->player->getStartLocation();
  this->startLocationPosition = Position(this->startLocationTilePosition.x() * TILE_SIZE, this->startLocationTilePosition.y() * TILE_SIZE);
  this->baseRegionCenter = BWTA::getRegion(this->player->getStartLocation())->getCenter();

  this->targetBuilding = NULL;
  this->targetTilePosition = TilePositions::None;
  this->targetRegionCenter = Positions::None;
  
  this->canSendAttackBuilder = false;
  this->attackBuilderReady = false;

  this->currentFrame = 0;
  this->lastActionFrame = 0;
  this->lastActionElapsedFrameCount = 0;

  this->lastAvailableActions = new int[Actions::count];
  this->lastAvailableActionCount = 0;

  this->isWaiting = false;
  this->enabledActions = new bool[Actions::count];

  this->accumulatedReward = 0;

  // load agent
  ifstream agentInput("bwapi-data/macroAIdata/agent.txt");
  if (agentInput) {
    this->agent = new RLAgent(&agentInput);
  } else {
    this->agent = new RLAgent(Actions::count, StateFields::count, AgentSettings::HiddenNeuronCount);
    this->agent->explorationRate = AgentSettings::ExplorationRate;
    this->agent->discountFactor = AgentSettings::DiscountFactor;
    this->agent->timeCoeficient = AgentSettings::TimeCoeficient;
    this->agent->setLearningRate(AgentSettings::LearningRate);
  }
  agentInput.close();

  this->isDraw = false;

  this->actionHistory.clear();
  this->actionHistory.reserve(500);

  this->stateHistory.clear();
  this->stateHistory.reserve(500);

  this->ignore = false;
}

void MacroAIModule::onEnd(bool isWinner) {
  if (!this->ignore) {
    // determine game outcome
    string outcome;
    double reward = 0;

    if (this->isDraw) {
      bool playerBeaten = true;
      for (set<Unit*>::const_iterator itUnit = this->player->getUnits().begin(); itUnit != this->player->getUnits().end(); ++itUnit) {
        if ((*itUnit)->getType().isWorker() || ((*itUnit)->getType().isBuilding() && (*itUnit)->getType().canProduce())) {
          playerBeaten = false;
          break;
        }
      }

      bool enemyPlayerBeaten = true;
      for (set<Unit*>::const_iterator itUnit = this->enemyPlayer->getUnits().begin(); itUnit != this->enemyPlayer->getUnits().end(); ++itUnit) {
        if ((*itUnit)->getType().isWorker() || ((*itUnit)->getType().isBuilding() && (*itUnit)->getType().canProduce())) {
          enemyPlayerBeaten = false;
          break;
        }
      }

      if (!playerBeaten && enemyPlayerBeaten) {
        outcome = "Good Draw";
        reward = 2;
      } else if (playerBeaten) {
        outcome = "Bad Draw";
        reward = -4;
      } else {
        outcome = "Draw";
        reward = -2;
      }
    } else if (isWinner) {
      outcome = "Victory";
      reward = 2;
    } else {
      outcome = "Defeat";
      reward = -4;
    }
    reward *= pow(this->agent->discountFactor, this->lastActionElapsedFrameCount);
    
    this->agent->initializeTurn();
    this->fillState();
    this->agent->learnFinal(this->lastActionElapsedFrameCount, reward);

    // save outcome
    ofstream outcomeOutput("bwapi-data/macroAIdata/results.txt", ios::out | ios::app);
    outcomeOutput << outcome << " : " << this->currentFrame << endl;
    outcomeOutput.close();

    // save game details
    ofstream detailsOutput("bwapi-data/macroAIdata/details.txt", ios::out | ios::app);
    detailsOutput << "[BEGIN GAME]" << endl;
    detailsOutput << setprecision(4);
    

    for (int i = 0; i < (int)this->actionHistory.size(); i++) {
      detailsOutput << setw(5) << setfill('0') << this->actionHistory[i].first << ":" << ActionNames[this->actionHistory[i].second] << endl;
      for (int j = 0; j < StateFields::count - 1; j++) {
        detailsOutput << this->stateHistory[i][j] << ",";
      }
      detailsOutput << this->stateHistory[i][StateFields::count - 1] << endl;
    }
    detailsOutput << setw(5) << setfill('0') << this->currentFrame << ":" << outcome << endl;
    for (int j = 0; j < StateFields::count - 1; j++) {
      detailsOutput << this->agent->state->get(j) << ",";
    }
    detailsOutput << this->agent->state->get(StateFields::count - 1) << endl;

    detailsOutput << "[END GAME]" << endl << endl;
    detailsOutput.close();

    // save agent
    ofstream agentOutput("bwapi-data/macroAIdata/agent.txt");
    agentOutput << setprecision(10);
    this->agent->save(&agentOutput);
    agentOutput.close();
  }

  delete[] this->lastAvailableActions;
  delete[] this->enabledActions;
  delete this->agent;
}

void MacroAIModule::onSendText(string text) {
  if ((text == "ignore") || (text == "quit")) {
    this->ignore = true;
    Broodwar->setGUI(true);
    Broodwar->setLocalSpeed(40);
    Broodwar->printf("This game will not be recorded ... you can now safely exit");
  }
}

void MacroAIModule::manageIncompleteUnits() {
  for (list<Unit*>::iterator itUnit = this->incompleteUnits.begin(); itUnit != this->incompleteUnits.end();) {
    if (!(*itUnit)->exists()) {
      itUnit = this->incompleteUnits.erase(itUnit);
      continue;
    }

    if ((*itUnit)->isCompleted()) {
      int unitTypeID = (*itUnit)->getType().getID();

      if (unitTypeID == UnitTypeIDs::ProtossProbe) {
        this->probes.insert(pair<int, Unit*>((*itUnit)->getID(), *itUnit));
      } else if (unitTypeID == UnitTypeIDs::ProtossPylon) {
        // add new buildable tiles
        int originX = (*itUnit)->getTilePosition().x();
        int originY = (*itUnit)->getTilePosition().y();

        for (int x = -PYLON_X_COVERAGE; x <= PYLON_X_COVERAGE; x++) {
          for (int y = -PYLON_Y_COVERAGE; y <= PYLON_Y_COVERAGE; y++) {
            TilePosition tile(originX + x, originY + y);
            BWTA::Region* tileRegion = BWTA::getRegion(tile);
            if ((tileRegion != NULL) && Broodwar->hasPower(tile) && Broodwar->isBuildable(tile, true)) {
              if (this->baseRegionCenter == tileRegion->getCenter()) {
                this->buildableBaseTiles.insert(tile);
                this->canBuildForge = true;
              } else if (this->targetRegionCenter == tileRegion->getCenter()) {
                this->buildableHostileTiles.insert(tile);
                this->canBuildAttackCannon = true;
              }
            }
          }
        }
      } else if (unitTypeID == UnitTypeIDs::ProtossPhotonCannon) {
        this->attackCannons.insert(pair<int, Unit*>((*itUnit)->getID(), *itUnit));

        // add new defended tiles
        int originX = (*itUnit)->getTilePosition().x();
        int originY = (*itUnit)->getTilePosition().y();

        for (int x = -this->weakDefenseDistance; x <= this->weakDefenseDistance; x++) {
          for (int y = -this->weakDefenseDistance; y <= this->weakDefenseDistance; y++) {
            TilePosition tile(originX + x, originY + y);
            BWTA::Region* tileRegion = BWTA::getRegion(tile);
            if (Broodwar->isBuildable(tile) && (tileRegion != NULL) && (this->targetRegionCenter == tileRegion->getCenter())) {
              if ((abs(x) <= this->strongDefenseDistance) && (abs(y) <= this->strongDefenseDistance)) {
                map<TilePosition, int>::iterator itDefTile = this->stronglyDefendedTiles.find(tile);
                if (itDefTile == this->stronglyDefendedTiles.end()) {
                  this->stronglyDefendedTiles.insert(pair<TilePosition, int>(tile, 1));
                  this->canBuildAttackPylon = true;
                } else {
                  itDefTile->second++;
                }
              } else {
                map<TilePosition, int>::iterator itDefTile = this->weaklyDefendedTiles.find(tile);
                if (itDefTile == this->weaklyDefendedTiles.end()) {
                  this->weaklyDefendedTiles.insert(pair<TilePosition, int>(tile, 1));
                } else {
                  itDefTile->second++;
                }
              }
            }
          }
        }
      }

      itUnit = this->incompleteUnits.erase(itUnit);
      continue;
    }

    ++itUnit;
  }
}

void MacroAIModule::manageBaseBuilderProbe() {
  if (this->baseBuilderProbe != NULL) {
    if (!this->baseBuilderProbe->exists()) {
      this->baseBuildInProgress = false;
      this->baseBuilderProbe = NULL;
    } else if (this->baseBuildInProgress && (this->lastActionElapsedFrameCount >= ORDER_DELAY)) {
      if (this->baseBuilderProbe->isIdle() || this->baseBuilderProbe->isGatheringMinerals()) {
        this->baseBuildInProgress = false;
      }
      if (this->lastActionElapsedFrameCount > this->buildTimeout) {
        this->baseBuildInProgress = false;
        Unit* closestMineral = this->getNearestMineralToNexus();
        if (closestMineral != NULL) {
          this->baseBuilderProbe->gather(closestMineral);
        }
        this->baseBuilderProbe = NULL;
      }
    }
  }
  if ((this->baseBuilderProbe == NULL) && !this->probes.empty()) {
    map<int, Unit*>::iterator itProbe = this->probes.begin();
    this->baseBuilderProbe = itProbe->second;
    this->probes.erase(itProbe);
  }
}

void MacroAIModule::manageAttackBuilderProbe() {
  if (this->attackBuilderProbe != NULL) {
    if (!this->attackBuilderProbe->exists()) {
      this->attackBuildInProgress = false;
      this->attackBuilderReady = false;
      this->attackBuilderProbe = NULL;
    } else if (this->attackBuildInProgress && (this->lastActionElapsedFrameCount >= ORDER_DELAY)) {
      if (this->attackBuilderProbe->isIdle() || this->attackBuilderProbe->isGatheringMinerals()) {
        this->attackBuildInProgress = false;
      }
      if (this->lastActionElapsedFrameCount > this->buildTimeout) {
        this->attackBuildInProgress = false;
        Unit* closestMineral = this->getNearestMineralToNexus();
        if (closestMineral != NULL) {
          this->attackBuilderProbe->gather(closestMineral);
        }
        this->attackBuilderReady = false;
        this->attackBuilderProbe = NULL;
      }
    }
  }
  if ((this->attackBuilderProbe == NULL) && !this->probes.empty()) {
    map<int, Unit*>::iterator itProbe = this->probes.begin();
    this->attackBuilderProbe = itProbe->second;
    this->probes.erase(itProbe);
  }
}

void MacroAIModule::issueOrdersToUnits() {
  // manage idle base builder probe
  if ((this->baseBuilderProbe != NULL) && !this->baseBuildInProgress && !this->baseBuilderProbe->isGatheringMinerals()) {
    Unit* closestMineral = this->getNearestMineralToNexus();
    if (closestMineral != NULL) {
      this->baseBuilderProbe->gather(closestMineral);
    }
  }

  // manage nonready attack builder probe
  if (this->attackBuilderProbe != NULL) {
    if (this->canSendAttackBuilder) {
      if (!this->attackBuilderReady && !this->attackBuildInProgress && (this->targetRegionCenter != Positions::None)) {
        BWTA::Region* unitRegion = BWTA::getRegion(this->attackBuilderProbe->getTilePosition());
        if ((unitRegion == NULL) || (this->targetRegionCenter != unitRegion->getCenter())) {
          this->attackBuilderProbe->move(this->targetRegionCenter);
        } else {
          this->attackBuilderProbe->stop();
          this->attackBuilderReady = true;
        }
      }
    } else if (!this->attackBuilderProbe->isGatheringMinerals()) {
      Unit* closestMineral = this->getNearestMineralToNexus();
      if (closestMineral != NULL) {
        this->attackBuilderProbe->gather(closestMineral);
      }
    }
  }

  // manage probes
  for (map<int, Unit*>::iterator itProbe = this->probes.begin(); itProbe != this->probes.end(); ++itProbe) {
    if (!itProbe->second->isGatheringMinerals()) {
      Unit* closestMineral = this->getNearestMineralToNexus();
      if (closestMineral != NULL) {
        itProbe->second->gather(closestMineral);
      }
    }
  }
}

void MacroAIModule::fillState() {
  // extract data from player units
  int unusedSupply = -this->player->supplyUsed();
  int probeCount = 0;
  
  int nexusCount = 0;
  int forgeCount = 0;
  int attackPylonCount = this->basePylon == NULL ? 0 : -1;
  int attackCannonCount = 0;

  for (set<Unit*>::const_iterator itUnit = this->player->getUnits().begin(); itUnit != this->player->getUnits().end(); ++itUnit) {
    int unitTypeID = (*itUnit)->getType().getID();
    if (unitTypeID == UnitTypeIDs::ProtossProbe) {
      probeCount++;
    } else if (unitTypeID == UnitTypeIDs::ProtossPylon) {
      unusedSupply += UnitTypes::Protoss_Pylon.supplyProvided();
      attackPylonCount++;
    } else if (unitTypeID == UnitTypeIDs::ProtossPhotonCannon) {
      attackCannonCount++;
    } else if (unitTypeID == UnitTypeIDs::ProtossNexus) {
      unusedSupply += UnitTypes::Protoss_Nexus.supplyProvided();
      nexusCount++;
    } else if (unitTypeID == UnitTypeIDs::ProtossForge) {
      forgeCount++;
    }
  }

  this->agent->state->set(StateFields::UnusedSupply, sig((0.4 * (double)(unusedSupply >> 1)) - 2.5));  
  this->agent->state->set(StateFields::ProbeCount, sig((0.375 * (double)probeCount) - 3.0));

  this->agent->state->set(StateFields::NexusCount, (double)(nexusCount > 0));
  this->agent->state->set(StateFields::ForgeCount, (double)(forgeCount > 0));
  double cannonPylonRatio = attackPylonCount == 0 ? 0 : ((double)attackCannonCount / attackPylonCount);
  this->agent->state->set(StateFields::CannonPylonRatio, tanh(0.3 * (double)cannonPylonRatio));

  // extract data from enemy player units
  int enemyWorkerCount = 0;
  int enemyTurretCount = 0;
  int enemyLowLevelProducerCount = 0;
  int enemyHighLevelProducerCount = 0;
  int enemyTechBuildingCount = 0;
  int enemyLowLevelUnitCount = 0;
  int enemyHightLevelUnitCount = 0;

  for (set<Unit*>::const_iterator itUnit = this->enemyPlayer->getUnits().begin(); itUnit != this->enemyPlayer->getUnits().end(); ++itUnit) {
    UnitType unitType = (*itUnit)->getType();
    int unitTypeID = unitType.getID();
    if (unitType.isWorker()) {
      enemyWorkerCount++;
    } else if (unitType.isBuilding()) {
      if (UnitTypeClasses::turretIDs.count(unitTypeID)) {
        enemyTurretCount++;
      } else if (UnitTypeClasses::lowLevelProducerIDs.count(unitTypeID)) {
        enemyLowLevelProducerCount++;
      } else if (UnitTypeClasses::highLevelProducerIDs.count(unitTypeID)) {
        enemyHighLevelProducerCount++;
      } else if (UnitTypeClasses::techBuildingIDs.count(unitTypeID)) {
        enemyTechBuildingCount++;
      }
    } else {
      if (UnitTypeClasses::lowLevelUnitIDs.count(unitTypeID)) {
        enemyLowLevelUnitCount++;
      } else if (UnitTypeClasses::hightLevelUnitIDs.count(unitTypeID)) {
        enemyHightLevelUnitCount++;
      }
    }
  }

  this->agent->state->set(StateFields::EnemyWorkerCount, sig((0.375 * (double)enemyWorkerCount) - 3.75));
  this->agent->state->set(StateFields::EnemyLowLevelUnitCount, tanh(0.175 * (double)enemyLowLevelUnitCount));
  this->agent->state->set(StateFields::EnemyHightLevelUnitCount, tanh(0.35 * (double)enemyHightLevelUnitCount));

  this->agent->state->set(StateFields::EnemyTurretCount, tanh(0.37 * (double)enemyTurretCount));
  this->agent->state->set(StateFields::EnemyLowLevelProducerCount, tanh(0.37 * (double)enemyLowLevelProducerCount));
  this->agent->state->set(StateFields::EnemyHighLevelProducerCount, tanh(0.37 * (double)enemyHighLevelProducerCount));
  this->agent->state->set(StateFields::EnemyTechBuildingCount, tanh(0.275 * (double)enemyTechBuildingCount));
}


void MacroAIModule::onFrame() {
  this->currentFrame = Broodwar->getFrameCount();
  this->lastActionElapsedFrameCount = currentFrame - this->lastActionFrame;
  
  // game timeout
  if (currentFrame > this->gameTimeoutFrame) {
    this->isDraw = true;
    Broodwar->restartGame();
  }

  // check incomplete units
  if (!this->incompleteUnits.empty()) {
    this->manageIncompleteUnits();
  }  

  // manage builder probe
  this->manageBaseBuilderProbe();
  this->manageAttackBuilderProbe();

  // issue orders to units
  if (!((currentFrame + 1) % ORDER_DELAY)) {
    this->issueOrdersToUnits();
  }

  // designate a new target if needed
  if ((this->attackBuilderProbe != NULL) && ((this->targetBuilding == NULL) || !this->targetBuilding->exists())) {
    this->chooseAttackTarget();
  }

  if (!this->baseBuildInProgress && !this->attackBuildInProgress) {
    // evaluate actions
    this->enabledActions[Actions::TrainProbe] =
      Broodwar->canMake(NULL, UnitTypes::Protoss_Probe);
    this->enabledActions[Actions::BuildBasePylon] =
      this->canBuildBasePylon &&
      (this->forge == NULL) &&
      (this->basePylon == NULL) &&
      (this->baseBuilderProbe != NULL) &&
      Broodwar->canMake(NULL, UnitTypes::Protoss_Pylon);    
    this->enabledActions[Actions::BuildForge] =
      this->canBuildForge &&
      (this->forge == NULL) &&
      (this->baseBuilderProbe != NULL) &&
      Broodwar->canMake(NULL, UnitTypes::Protoss_Forge);
    this->enabledActions[Actions::BuildAttackPylon] =
      this->attackBuilderReady &&
      this->canBuildAttackPylon &&
      (this->attackPylons.empty() || !this->attackCannons.empty()) &&
      Broodwar->canMake(NULL, UnitTypes::Protoss_Pylon);
    this->enabledActions[Actions::BuildAttackCannon] =
      this->attackBuilderReady &&
      this->canBuildAttackCannon &&
      !this->buildableHostileTiles.empty() &&
      Broodwar->canMake(NULL, UnitTypes::Protoss_Photon_Cannon);
    this->enabledActions[Actions::Wait] =
      this->player->minerals() < this->highestMineralActionPrice;

    // choose training nexus
    Unit* trainingNexus = NULL;
    if (this->enabledActions[Actions::TrainProbe]) {
      if ((this->nexus != NULL) && !this->nexus->isTraining()) {
        trainingNexus = this->nexus;
      }
      this->enabledActions[Actions::TrainProbe] = trainingNexus != NULL;
    }

    // determine available actions    
    this->agent->availableActionCount = 0;
    for (int i = 0; i < Actions::count; i++) {
      if (this->enabledActions[i]) {
        this->agent->availableActions[this->agent->availableActionCount] = i;
        this->agent->availableActionCount++;
      }
    }

    // reevaluate wait action
    if (this->enabledActions[Actions::Wait] && ((this->agent->availableActionCount == 1) || (this->agent->availableActionCount == Actions::count))) {
      this->enabledActions[Actions::Wait] = false;
      this->agent->availableActionCount--;
    }

    if (this->agent->availableActionCount > 0) {
      // evaluate is waiting
      if (this->isWaiting) {
        this->isWaiting = this->lastActionElapsedFrameCount < this->maxWaitFrameCount;
        if (this->isWaiting) {
          this->isWaiting = this->lastAvailableActionCount == this->agent->availableActionCount;
          if (this->isWaiting) {
            for (int i = 0; this->isWaiting && (i < this->lastAvailableActionCount); i++) {
              this->isWaiting = this->agent->availableActions[i] == this->lastAvailableActions[i];
            }
          }
        }
      }

      if (!this->isWaiting) {
        this->agent->initializeTurn();
        this->fillState();

        this->agent->calculateValues();
        this->agent->learn(this->lastActionElapsedFrameCount, this->accumulatedReward);
        this->agent->chooseAction();

        if (this->agent->action == Actions::TrainProbe) {
          trainingNexus->train(UnitTypes::Protoss_Probe);
        } else if (this->agent->action == Actions::BuildBasePylon) {
          this->baseBuildInProgress = this->baseBuilderProbe->build(this->getBasePylonBuildTile(), UnitTypes::Protoss_Pylon);
        } else if (this->agent->action == Actions::BuildForge) {
          this->baseBuildInProgress = this->baseBuilderProbe->build(this->getForgeBuildTile(), UnitTypes::Protoss_Forge);
        } else if (this->agent->action == Actions::BuildAttackPylon) {
          this->attackBuildInProgress = this->attackBuilderProbe->build(this->getAttackPylonBuildTile(), UnitTypes::Protoss_Pylon);
        } else if (this->agent->action == Actions::BuildAttackCannon) {
          this->attackBuildInProgress = this->attackBuilderProbe->build(this->getAttackCannonBuildTile(), UnitTypes::Protoss_Photon_Cannon);
        } else if (this->agent->action == Actions::Wait) {
          this->isWaiting = true;
        }
          
        this->lastActionFrame = this->currentFrame;
        for (int i = 0; i < this->agent->availableActionCount; i++) {
          this->lastAvailableActions[i] = this->agent->availableActions[i];
        }
        this->lastAvailableActionCount = this->agent->availableActionCount;

        this->accumulatedReward = 0;

        this->actionHistory.push_back(pair<int, int>(this->currentFrame, this->agent->action));

        vector<double> stateData;
        for (int i = 0; i < StateFields::count; i++) {
          stateData.push_back(this->agent->state->get(i));
        }
        this->stateHistory.push_back(stateData);
      }
    }
  }
}

void MacroAIModule::onUnitCreate(Unit* unit) {
  if (unit->getPlayer()->isNeutral()) {
    return;
  }

  // add to incomplete units
  if (unit->getPlayer()->getID() == this->player->getID()) {
    this->incompleteUnits.push_back(unit);
  }

  if (unit->getType().isBuilding()) {
    Position unitRegionCenter = BWTA::getRegion(unit->getTilePosition())->getCenter();

    // register precursor buildings
    if (unit->getPlayer()->getID() == this->player->getID()) {
      if (unit->getType().getID() == UnitTypeIDs::ProtossForge) {
        this->forge = unit;
      } else if (unit->getType().getID() == UnitTypeIDs::ProtossPylon) {
        if (unitRegionCenter == this->baseRegionCenter) {
          this->basePylon = unit;
          this->canSendAttackBuilder = true;
        } else if (unitRegionCenter == this->targetRegionCenter) {
          this->attackPylons.insert(pair<int, Unit*>(unit->getID(), unit));
        }
      }
    }

    // remove occupied build tiles
    int originX = unit->getTilePosition().x();
    int originY = unit->getTilePosition().y();
    int buildWidth = unit->getType().tileWidth();
    int buildHeight = unit->getType().tileHeight();

    if (unitRegionCenter == this->baseRegionCenter) {
      for (int x = 0; x < buildWidth; x++) {
        for (int y = 0; y < buildHeight; y++) {
          this->buildableBaseTiles.erase(TilePosition(originX + x, originY + y));
        }
      }
    } else if (unitRegionCenter == this->targetRegionCenter) {
      for (int x = 0; x < buildWidth; x++) {
        for (int y = 0; y < buildHeight; y++) {
          this->buildableHostileTiles.erase(TilePosition(originX + x, originY + y));
        }
      }
    }
  }
}

void MacroAIModule::manageLostUnit(Unit* unit) {
  int unitTypeID = unit->getType().getID();

  if (unitTypeID == UnitTypeIDs::ProtossProbe) {
    this->probes.erase(unit->getID());
  } else if (unitTypeID == UnitTypeIDs::ProtossPylon) {
    // remove lost buildable tiles
    int originX = unit->getTilePosition().x();
    int originY = unit->getTilePosition().y();

    Position unitRegionCenter = BWTA::getRegion(unit->getTilePosition())->getCenter();

    if (unitRegionCenter == this->baseRegionCenter) {
      this->basePylon = NULL;

      for (int x = -PYLON_X_COVERAGE; x <= PYLON_X_COVERAGE; x++) {
        for (int y = -PYLON_Y_COVERAGE; y <= PYLON_Y_COVERAGE; y++) {
          TilePosition tile(originX + x, originY + y);
          if (!Broodwar->hasPower(tile)) {
            this->buildableBaseTiles.erase(tile);
          }
        }
      }
    } else if (unitRegionCenter == this->targetRegionCenter) {
      this->attackPylons.erase(unit->getID());

      for (int x = -PYLON_X_COVERAGE; x <= PYLON_X_COVERAGE; x++) {
        for (int y = -PYLON_Y_COVERAGE; y <= PYLON_Y_COVERAGE; y++) {
          TilePosition tile(originX + x, originY + y);
          if (!Broodwar->hasPower(tile)) {
            this->buildableHostileTiles.erase(tile);
          }
        }
      }
    }
  } else if (unitTypeID == UnitTypeIDs::ProtossForge) {
    this->forge = NULL;
  } else if (unitTypeID == UnitTypeIDs::ProtossNexus) {
    this->nexus = NULL;
  } else if (unitTypeID == UnitTypeIDs::ProtossPhotonCannon) {
    this->attackCannons.erase(unit->getID());

    // remove lost defended tiles
    int originX = unit->getTilePosition().x();
    int originY = unit->getTilePosition().y();

    for (int x = -this->weakDefenseDistance; x <= this->weakDefenseDistance; x++) {
      for (int y = -this->weakDefenseDistance; y <= this->weakDefenseDistance; y++) {
        TilePosition tile(originX + x, originY + y);
        BWTA::Region* tileRegion = BWTA::getRegion(tile);
        if (Broodwar->isBuildable(tile) && (tileRegion != NULL) && (this->targetRegionCenter == tileRegion->getCenter())) {
          if ((abs(x) <= this->strongDefenseDistance) && (abs(y) <= this->strongDefenseDistance)) {
            map<TilePosition, int>::iterator itDefTile = this->stronglyDefendedTiles.find(tile);
            if (itDefTile != this->stronglyDefendedTiles.end()) {
              itDefTile->second--;
              if (!itDefTile->second) {
                this->stronglyDefendedTiles.erase(itDefTile);
              }
            }
          } else {
            map<TilePosition, int>::iterator itDefTile = this->weaklyDefendedTiles.find(tile);
            if (itDefTile != this->weaklyDefendedTiles.end()) {
              itDefTile->second--;
              if (!itDefTile->second) {
                this->weaklyDefendedTiles.erase(itDefTile);
              }
            }
          }
        }
      }
    }
  }
}


void MacroAIModule::onUnitDestroy(Unit* unit) {
  if (unit->getPlayer()->isNeutral()) {
    return;
  }
  
  // manage lost unit
  if (unit->getPlayer()->getID() == this->player->getID()) {
    this->manageLostUnit(unit);
  }

  // add freed buildable tiles
  if (unit->getType().isBuilding()) {
    if (unit->getPlayer()->getID() == this->enemyPlayer->getID()) {
      this->accumulatedReward += 0.1 * pow(this->agent->discountFactor, this->lastActionElapsedFrameCount);
    }

    int originX = unit->getTilePosition().x();
    int originY = unit->getTilePosition().y();
    int buildWidth = unit->getType().tileWidth();
    int buildHeight = unit->getType().tileHeight();

    for (int x = 0; x < buildWidth; x++) {
      for (int y = 0; y < buildHeight; y++) {
        TilePosition tile(originX + x, originY + y);
        BWTA::Region* tileRegion = BWTA::getRegion(tile);
        if ((tileRegion != NULL) && Broodwar->hasPower(tile)) {
          if (this->baseRegionCenter == tileRegion->getCenter()) {
            this->buildableBaseTiles.insert(tile);
          } else if (this->targetRegionCenter == tileRegion->getCenter()) {
            this->buildableHostileTiles.insert(tile);
          }
        }
      }
    }
  }
}

void MacroAIModule::onUnitRenegade(Unit* unit) {
  if (unit->getPlayer()->isNeutral()) {
    return;
  }

  // manage lost unit
  if ((unit->getPlayer()->getID() == this->enemyPlayer->getID())) {
    this->manageLostUnit(unit);
  }
}

Unit* MacroAIModule::getNearestMineralToNexus() {
  Unit* closestMineral = NULL;

  if (this->nexus != NULL) {
    int closestDistance = 0;  
    for (set<Unit*>::iterator itMineral = Broodwar->getMinerals().begin(); itMineral != Broodwar->getMinerals().end(); ++itMineral) {
      int distance = this->nexus->getDistance(*itMineral);
      if ((closestMineral == NULL) || (distance < closestDistance)) {
        closestMineral = *itMineral;
        closestDistance = distance;
      }
    }
  }

  return closestMineral;
}

void MacroAIModule::chooseAttackTarget() {
  this->targetBuilding = NULL;
  int closestDistance = 0;

  Position attackBuilderPosition(this->attackBuilderProbe->getPosition());

  // search for closest resource depot
  for (set<Unit*>::const_iterator itUnit = this->enemyPlayer->getUnits().begin(); itUnit != this->enemyPlayer->getUnits().end(); ++itUnit) {
    if ((*itUnit)->getType().isResourceDepot()) {
      int distance = attackBuilderPosition.getApproxDistance((*itUnit)->getPosition());
      if ((this->targetBuilding == NULL) || (distance < closestDistance)) {
        closestDistance = distance;
        this->targetBuilding = *itUnit;
      }
    }
  }

  // search for closest building
  if (this->targetBuilding == NULL) {
    for (set<Unit*>::const_iterator itUnit = this->enemyPlayer->getUnits().begin(); itUnit != this->enemyPlayer->getUnits().end(); ++itUnit) {
      if ((*itUnit)->getType().isBuilding()) {
        int distance = attackBuilderPosition.getApproxDistance((*itUnit)->getPosition());
        if ((this->targetBuilding == NULL) || (distance < closestDistance)) {
          closestDistance = distance;
          this->targetBuilding = *itUnit;
        }
      }
    }
  }

  if (this->targetBuilding != NULL) {
    this->targetTilePosition = this->targetBuilding->getTilePosition();
    Position newTargetRegionCenter(BWTA::getRegion(this->targetTilePosition)->getCenter());
    if (newTargetRegionCenter != this->targetRegionCenter) {
      this->targetRegionCenter = newTargetRegionCenter;
      
      this->attackBuilderReady = false;
      this->attackPylons.clear();
      this->attackCannons.clear();

      this->buildableHostileTiles.clear();
      this->stronglyDefendedTiles.clear();
      this->weaklyDefendedTiles.clear();

      this->hostileAreaTiles.clear();
      this->hostileAreaSize = this->restrictedFirstPylonArea + this->areaExpandStep;
      this->canBuildAttackPylon = true;
    }
  }
}

TilePosition MacroAIModule::getBasePylonBuildTile() {
  while (this->canBuildBasePylon) {
    if (this->baseAreaTiles.empty()) {
      this->canBuildBasePylon = false;
      
      int enabledSection = this->baseAreaSize - this->areaExpandStep;
      this->baseAreaTiles.reserve(this->areaExpandStep * ((this->baseAreaSize << 1) - this->areaExpandStep));

      int originX = this->startLocationTilePosition.x();
      int originY = this->startLocationTilePosition.y();

      for (int x = -this->baseAreaSize; x <= this->baseAreaSize; x++) {
		    for (int y = -this->baseAreaSize; y <= this->baseAreaSize; y++) {
          if ((abs(x) >= enabledSection) || (abs(y) >= enabledSection)) {
            TilePosition tile(originX + x, originY + y);
            BWTA::Region* tileRegion = BWTA::getRegion(tile);
            if (Broodwar->isBuildable(tile) && (tileRegion != NULL) && (this->baseRegionCenter == tileRegion->getCenter())) {
              this->canBuildBasePylon = true;
              this->baseAreaTiles.push_back(tile);
            }
          }
		    }
	    }

      random_shuffle(this->baseAreaTiles.begin(), this->baseAreaTiles.end());
    }

    for (int i = 0; i < (int)this->baseAreaTiles.size(); i++) {
      if (Broodwar->canBuildHere(NULL, this->baseAreaTiles[i], UnitTypes::Protoss_Pylon, true)) {
		    return this->baseAreaTiles[i];
	    }
	  }

    if (this->canBuildBasePylon) {
      this->baseAreaTiles.clear();
      this->baseAreaSize += this->areaExpandStep;
    }
  }

  return TilePositions::None;
}

TilePosition MacroAIModule::getForgeBuildTile() {
  for (set<TilePosition>::iterator itTile = this->buildableBaseTiles.begin(); itTile != this->buildableBaseTiles.end(); ++itTile) {
    if (Broodwar->canBuildHere(NULL, *itTile, UnitTypes::Protoss_Forge, true)) {
      return *itTile;
    }
  }
  this->canBuildForge = false;
  return TilePositions::None;
}
 
bool MacroAIModule::isPositionConcealed(Position position) {
  for (set<Unit*>::const_iterator itUnit = this->enemyPlayer->getUnits().begin(); itUnit != this->enemyPlayer->getUnits().end(); ++itUnit) {
    if (position.getApproxDistance((*itUnit)->getPosition()) < this->positionConcealedDistance) {
      return false;
    } 
  }
  return true;
}

int MacroAIModule::tileDistanceValue(TilePosition tile) {
  int result = 0;
  for (set<Unit*>::const_iterator itUnit = this->enemyPlayer->getUnits().begin(); itUnit != this->enemyPlayer->getUnits().end(); ++itUnit) {
    if ((*itUnit)->getType().isBuilding()) {
      result += (int)BWTA::getGroundDistance(tile, (*itUnit)->getTilePosition());
    }
  }
  return result;
}

TilePosition MacroAIModule::getAttackPylonBuildTile() {
  int bestValue = 0;
  TilePosition bestTilePosition(TilePositions::None);

  if (!this->attackCannons.empty()) {
    // search for suitable defended position
    for (map<TilePosition, int>::iterator itTile = this->stronglyDefendedTiles.begin(); itTile != this->stronglyDefendedTiles.end(); ++itTile) {
      if (Broodwar->canBuildHere(NULL, itTile->first, UnitTypes::Protoss_Pylon, false)) {
        int value = (int)BWTA::getGroundDistance(itTile->first, this->targetTilePosition);
        if ((bestTilePosition == TilePositions::None) || (value < bestValue)) {
          bestValue = value;
          bestTilePosition = itTile->first;
        }
      }
    }
  } else {
    // search for suitable concealed position
    while (this->canBuildAttackPylon) {
      if (this->hostileAreaTiles.empty()) {
        this->canBuildAttackPylon = false;
        
        int enabledSection = this->hostileAreaSize - this->areaExpandStep;
        this->hostileAreaTiles.reserve(this->areaExpandStep * ((this->hostileAreaSize << 1) - this->areaExpandStep));

        int originX = this->targetTilePosition.x();
        int originY = this->targetTilePosition.y();

        for (int x = -this->hostileAreaSize; x <= this->hostileAreaSize; x++) {
		      for (int y = -this->hostileAreaSize; y <= this->hostileAreaSize; y++) {
            if ((abs(x) >= enabledSection) || (abs(y) >= enabledSection)) {
              TilePosition tile(originX + x, originY + y);
              BWTA::Region* tileRegion = BWTA::getRegion(tile);
              if (Broodwar->isBuildable(tile) && (tileRegion != NULL) && (this->targetRegionCenter == tileRegion->getCenter())) {
                this->canBuildAttackPylon = true;
                this->hostileAreaTiles.push_back(tile);
              }
            }
		      }
	      }

        random_shuffle(this->hostileAreaTiles.begin(), this->hostileAreaTiles.end());
      }

      for (int i = 0; i < (int)this->hostileAreaTiles.size(); i++) {
        if (Broodwar->canBuildHere(NULL, this->hostileAreaTiles[i], UnitTypes::Protoss_Pylon, true)) {
          if (this->isPositionConcealed(Position(this->hostileAreaTiles[i].x() * TILE_SIZE, this->hostileAreaTiles[i].y() * TILE_SIZE))) {
		        int value = (int)BWTA::getGroundDistance(this->hostileAreaTiles[i], this->targetTilePosition);
            if ((bestTilePosition == TilePositions::None) || (value < bestValue)) {
              bestValue = value;
              bestTilePosition = this->hostileAreaTiles[i];
            }
          }
	      }
	    }

      if (bestTilePosition != TilePositions::None) {
        return bestTilePosition;
      }

      if (this->canBuildAttackPylon)  {
        this->hostileAreaTiles.clear();
        this->hostileAreaSize += this->areaExpandStep;
      }
    }
  }

  if (bestTilePosition == TilePositions::None) {
    this->canBuildAttackPylon = false;
  }

  return bestTilePosition;
}

TilePosition MacroAIModule::getAttackCannonBuildTile() {
  int bestValue = 0;
  TilePosition bestTilePosition(TilePositions::None);

  // search for suitable weakly defended position
  for (map<TilePosition, int>::iterator itTile = this->weaklyDefendedTiles.begin(); itTile != this->weaklyDefendedTiles.end(); ++itTile) {
    if (Broodwar->canBuildHere(NULL, itTile->first, UnitTypes::Protoss_Photon_Cannon, false)) {
      int value = this->tileDistanceValue(itTile->first);
      if ((bestTilePosition == TilePositions::None) || (value < bestValue)) {
        bestValue = value;
        bestTilePosition = itTile->first;
      }
    }
  }

  if (bestTilePosition == TilePositions::None) {
    // search for safest buildable position
    for (set<TilePosition>::iterator itTile = this->buildableHostileTiles.begin(); itTile != this->buildableHostileTiles.end(); ++itTile) {
      if (Broodwar->canBuildHere(NULL, *itTile, UnitTypes::Protoss_Photon_Cannon, false)) {
        int value = this->tileDistanceValue(*itTile);
        if ((bestTilePosition == TilePositions::None) || (value > bestValue)) {
          bestValue = value;
          bestTilePosition = *itTile;
        }
      }
    }
  }

  if (bestTilePosition == TilePositions::None) {
    this->canBuildAttackCannon = false;
  }

  return bestTilePosition;
}