#include <cstdlib>
#include <fstream>
#include <iostream>
#include <cmath>

#include <RL/Matrix.h>
#include <RL/NeuronNetwork.h>

NeuronNetwork::NeuronNetwork(int inputNodeCount, int hiddenNodeCount, int outputNodeCount)
  : learningRate(1)
  , hiddenLayerWeights(new Matrix(inputNodeCount + 1, hiddenNodeCount))
  , outputLayerWeights(new Matrix(hiddenNodeCount + 1, outputNodeCount))
  , hiddenActivation(hiddenNodeCount)
  , outputActivation(outputNodeCount)
  , hiddenError(hiddenNodeCount)
  , outputError(outputNodeCount)
{
	double hiddenRandFactor = 1 / sqrt((double)inputNodeCount + 1);
	for (int i = 0; i < inputNodeCount + 1; i++) {
		for (int j = 0; j < hiddenNodeCount; j++) {
			this->hiddenLayerWeights->set(i, j, ((((double)rand() / RAND_MAX) * 2) - 1) * hiddenRandFactor);
		}
	}

  double outputRandFactor = 1 / sqrt((double)hiddenNodeCount + 1);
	for (int i = 0; i < hiddenNodeCount + 1; i++) {
		for (int j = 0; j < outputNodeCount; j++) {
			this->outputLayerWeights->set(i, j, ((((double)rand() / RAND_MAX) * 2) - 1) * outputRandFactor);
		}
	}
}

NeuronNetwork::NeuronNetwork(ifstream* input) {
	*input >> this->learningRate;

  this->hiddenLayerWeights = new Matrix(input);
  this->outputLayerWeights = new Matrix(input);
}

void NeuronNetwork::evaluate(Matrix* input, Matrix* output) {
  // duplicate input matrix for output
  output->extract(input);

  // add bias inputs and run through hidden layer
    // Y = f(X x W)
	output->extend(-1);
	output->multiply(this->hiddenLayerWeights);
	output->applySigmoid();

  // add bias inputs and run through output layer
    // Z = g(Y x V)
	output->extend(-1);
	output->multiply(this->outputLayerWeights);
}

void NeuronNetwork::learn(Matrix* input, Matrix* target) {
  // duplicate input matrix for hiddenActivations
	this->hiddenActivation.extract(input);

  // add bias inputs and run through hidden layer
    // Y = f(X x W)
	this->hiddenActivation.extend(-1);
	this->hiddenActivation.multiply(this->hiddenLayerWeights);
	this->hiddenActivation.applySigmoid();

  // duplicate hiddenActivations for output 
	this->outputActivation.extract(&this->hiddenActivation);

  // add bias inputs and run through output layer
    // Z = g(Y x V)
	this->outputActivation.extend(-1);
	this->outputActivation.multiply(this->outputLayerWeights);

  // determine output layer error
    // eZ = T - Z
	this->outputError.extract(target);
	this->outputError.subtract(&this->outputActivation);

  // use back propagation to calculate hidden layer error
    // eY = f'(Y) * (eZ x trans(V))
	this->hiddenError.extract(&this->hiddenActivation);
	this->hiddenError.applyDiffSigmoid();
	this->tempMatrix.extract(&this->outputError); 
	this->tempMatrix.multiplyTransposed(this->outputLayerWeights);
	this->hiddenError.scalarMultiply(&this->tempMatrix);

  // modify output layer weights
    // V = V + alfa * (trans(Y) x eZ)
	this->tempMatrix.extract(&this->hiddenActivation);
	this->tempMatrix.extend(-1);
	this->tempMatrix.transpose();
	this->tempMatrix.multiply(&this->outputError);
	this->tempMatrix.scalarMultiply(this->learningRate);
	this->outputLayerWeights->add(&this->tempMatrix);

  // modify hidden layer weights
    // W = W + alfa * (trans(X) x eY)
	this->tempMatrix.extract(input);
	this->tempMatrix.extend(-1);
	this->tempMatrix.transpose();
	this->tempMatrix.multiply(&this->hiddenError);
	this->tempMatrix.scalarMultiply(this->learningRate);
	this->hiddenLayerWeights->add(&this->tempMatrix);
}

void NeuronNetwork::save(ofstream* output) {
	*output << this->learningRate << endl;

	this->hiddenLayerWeights->save(output);
	this->outputLayerWeights->save(output);
}

NeuronNetwork::~NeuronNetwork() {
  delete this->hiddenLayerWeights;
  delete this->outputLayerWeights;
}