#include <vector>
#include <fstream>
#include <cmath>
#include <cassert>

#include <RL/Matrix.h>
#include <RL/NeuronNetwork.h>
#include <RL/RLAgent.h>

RLAgent::RLAgent(int newActionCount, int newStateDimensionCount, int newHiddenNeuronCount)
  : explorationRate(1)
  , discountFactor(0)
  , timeCoeficient(1)
  , state(new Matrix(newStateDimensionCount))
  , action(NULL_ACTION)
  , availableActions(new int[newActionCount])
  , availableActionCount(0)
  , lastState(new Matrix(newStateDimensionCount))
  , lastAction(NULL_ACTION)
  , bestAction(NULL_ACTION)
  , bestValue(0)
  , lastBestValue(0)
  , values(new double[newActionCount])
  , boltzmannValues(new double[newActionCount])
  , actionCount(newActionCount)
  , stateDimensionCount(newStateDimensionCount)
{
	for (int i = 0; i < newActionCount; i++) {
		this->actionNetworks.push_back(new NeuronNetwork(newStateDimensionCount, newHiddenNeuronCount, 1));
	}
}

RLAgent::RLAgent(ifstream* input)
  : action(NULL_ACTION)
  , availableActionCount(0)
  , lastAction(NULL_ACTION)
  , bestAction(NULL_ACTION)
  , bestValue(0)
  , lastBestValue(0)
{
  *input >> this->explorationRate;
  *input >> this->discountFactor;
  *input >> this->timeCoeficient;
	*input >> this->actionCount;
  *input >> this->stateDimensionCount;

  this->state = new Matrix(this->stateDimensionCount);
  this->availableActions = new int[this->actionCount];
  this->lastState = new Matrix(this->stateDimensionCount);
  this->values = new double[this->actionCount];
  this->boltzmannValues = new double[this->actionCount];

  for (int i = 0; i < this->actionCount; i++) {
		this->actionNetworks.push_back(new NeuronNetwork(input));
	}
}

void RLAgent::setLearningRate(double newLearningRate) {
	for (vector<NeuronNetwork*>::iterator it = this->actionNetworks.begin(); it != this->actionNetworks.end(); ++it) {
    (*it)->learningRate = newLearningRate;
	}
}

void RLAgent::initializeTurn() {
  Matrix* swapper = this->lastState;
  this->lastState = this->state;
  this->state = swapper;

  this->lastAction = this->action;
  this->action = NULL_ACTION;  

  this->bestAction = NULL_ACTION;

  this->lastBestValue = this->bestValue;
  this->bestValue = 0;
}

void RLAgent::calculateValues() {
  this->bestAction = NULL_ACTION;
  for (int i = 0; i < this->availableActionCount; i++) {
    this->actionNetworks[this->availableActions[i]]->evaluate(this->state, &this->valueMatrix);
    this->values[i] = this->valueMatrix.get();
    if ((this->bestAction == NULL_ACTION) || (this->values[i] > this->bestValue)) {
      this->bestAction = this->availableActions[i];
      this->bestValue = this->values[i];
    }
  }
}

void RLAgent::learn(int timeElapsed, double reward) {
  if (this->lastAction != NULL_ACTION) {
    // target = R + gamma^t * max(Q(s', a'))
    double targetValue = reward;
    targetValue += pow(this->discountFactor, timeElapsed) * this->bestValue;
#if LA_QLearning
    // target
#elif LA_AdvantageLearning
    // max(Q(s, a)) + (target - max(Q(s, a))) / t * K
    targetValue -= this->lastBestValue;
    targetValue /= timeElapsed * this->timeCoeficient;
    targetValue += this->lastBestValue;
#else
    assert(0);
#endif
    this->valueMatrix.set(targetValue);
    this->actionNetworks[this->lastAction]->learn(this->lastState, &this->valueMatrix);
  }
}

void RLAgent::learnFinal(int timeElapsed, double reward) {
  if (this->lastAction != NULL_ACTION) {
    // target = R
    double targetValue = reward;
#if LA_QLearning
    // target
#elif LA_AdvantageLearning
    // max(Q(s, a)) + (target - max(Q(s, a))) / t * K
    targetValue -= this->lastBestValue;
    targetValue /= timeElapsed * this->timeCoeficient;
    targetValue += this->lastBestValue;
#else
    assert(0);
#endif
    this->valueMatrix.set(targetValue);
    this->actionNetworks[this->lastAction]->learn(this->lastState, &this->valueMatrix);
  }
}

void RLAgent::chooseAction() {
#if EM_EpsylonGreedy
  if (this->explorationRate > ((double)rand() / (RAND_MAX + 1))) {
    this->action = this->availableActions[rand() % this->availableActionCount];
  }	else {
    this->action = this->bestAction;
  }
#elif EM_Boltzmann
  double boltzmannSum = 0;
  for (int i = 0; i < this->availableActionCount; i++) {
    this->boltzmannValues[i] = exp(this->values[i] / this->explorationRate);
    boltzmannSum += this->boltzmannValues[i];
  }

  double randValue = (double)rand() / (RAND_MAX + 1);
  for (int i = 0; i < this->availableActionCount; i++) {
    randValue -= this->boltzmannValues[i] / boltzmannSum;
    if (randValue < 0) {
      this->action = this->availableActions[i];
      return;
    }
  }
  if (this->action == -1) {
    this->action = this->availableActions[rand() % this->availableActionCount];
  }
#else
  assert(0);
#endif 
}

void RLAgent::save(ofstream* output) {
  *output << this->explorationRate << endl;
  *output << this->discountFactor << endl;
  *output << this->timeCoeficient << endl;
	*output << this->actionCount << endl;
  *output << this->stateDimensionCount << endl;

  for (vector<NeuronNetwork*>::iterator it = this->actionNetworks.begin(); it != this->actionNetworks.end(); ++it) {
		(*it)->save(output);
	}
}

RLAgent::~RLAgent() {
  delete this->state;
  delete[] this->availableActions;
  delete this->lastState;
  delete[] this->values;
  delete[] this->boltzmannValues;

  for (vector<NeuronNetwork*>::iterator it = this->actionNetworks.begin(); it != this->actionNetworks.end(); ++it) {
		delete *it;
	}
}